#include <string>
#include <iostream>

//#define _GNU_SOURCE
#include <unistd.h>
#include <getopt.h>

extern "C" {
#include "text.h"
};

using namespace std;

static void Adder(void *p, const char *s)
{
    *((string *)p) += s;
}
static string Phonemize(const string &s)
{
    string res;
    NRL(s.c_str(), s.size(), &res, Adder);
    return res;
}

/*
#define PHONE(nm,br,am,ex) {br,  "?"} nm ex
#include "/usr/src/muu/rsynth-2.0/phones.def"
*/

static bool ShowPhonemes = false;

static int Romaji(const string &english)
{
    /* Phonemes to romaji conversion table */
    static const char tab[][2][5] =
{
    {"@U", "o^"  }   /* OW "goat,shOW,nO"                               */
   ,{ "@", "a"   }   /* AX "About,commA,commOn"                         */
   ,{ "A", "a^"  }   /* AA "stARt,fAther"                               */
   ,{ "&", "a"   }   /* AE "trAp,bAd"                                   */
   ,{ "V", "a^"  }   /* AH "strUt,bUd,lOve"                             */
   ,{"Or", "oo"  }   /* AR "wARp" */
   ,{ "O", "o^"  }   /* AO "thOUGHt,lAW"                                */
   ,{"aU", "au"  }   /* AW "mOUth,nOW"                                  */
   ,{"aI", "ai"  }   /* AY "prIce,hIGH,trY"                             */
   ,{"e@", "e^"  }   /* EA "squARE,fAIR"                                */
   ,{"eI", "e^"  }   /* EY "fAce,dAy,stEAk"                             */
   ,{ "e", "e"   }   /* EH "drEss,bEd"                                  */
   ,{ "3", "a^"  }   /* ER "nURse,stIR,cOURage"                         */
   ,{"ir", "ia^" }   /* IA "nEAR,hERE,sErious"                          */
   ,{"i@", "ia"  }   /* EO "mEOw"                                       */
   ,{"I@", "ya"  }   /* YA */
   ,{ "I", "i"   }   /* IH "Intend,basIc,kIt,bId,hYmn"                  */
   ,{ "i", "i^"  }   /* IY "happY,radIation,glorIous,flEEce,sEA,machIne"*/
   ,{"0r", "a^"  }   /* ar*/
   ,{ "0", "o"   }   /* OH "lOt,Odd,wAsh"                               */
   ,{"oI", "oi"  }   /* OY "chOIce,bOY"                                 */
   ,{"U@", "ue^" }   /* UA "inflUence,sitUation,annUal,cURE,pOOR,jUry"  */
   ,{ "U", "u"   }   /* UH "fOOt,gOOd,pUt,stimUlus,edUcate"             */
   ,{ "u", "u^"  }   /* UW "gOOse,twO,blUE"                             */
   ,{"Z@", "jo"  }   /* viSIOn  */

   ,{ "b", "b"   }   /* B "Back,BuBBle,joB"                             */
   ,{"dZe","ze"  }
   ,{"dZ", "j"   }   /* JH "JuDGE,aGe,solDIer"                          */
   ,{ "d", "d"   }   /* D "Day,laDDer,oDD"                              */
   ,{"DI", "ti"  }   /* DI "THIs"                                       */
   ,{ "D", "z"   }   /* DH "THis,oTHer,smooTH"                          */
   ,{ "f", "f"   }   /* F "Fat,coFFee,rouGH,PHysics"                    */
   ,{ "g", "g"   }   /* G "Get,GiGGle,GHost"                            */
   ,{ "h", "h"   }   /* HH "Hot,WHole,beHind"                           */
   ,{"k&n","kan" }   /* CAN "vulCAN" */
   ,{"k&", "kya" }   /* CA "CAt,CAptain" */
   ,{ "k", "k"   }   /* K "Key,CoCK,sCHool"                             */
   ,{ "l", "r"   }   /* L "Light,vaLLey,feeL"                           */
   ,{ "m", "m"   }   /* M "More,haMMer,suM"                             */
   ,{ "n", "n"   }   /* N "Nice,KNow,fuNNy,suN"                         */
   ,{"Ng", "n"   }   /* NG "coNGratulations"                            */
   ,{ "N", "n"   }   /* NG "riNG,loNG,thaNks,suNG"                      */
   ,{ "p", "p"   }   /* P "Pen,coPy,haPPen"                             */
   ,{ "r", "r"   }   /* R "Right,soRRy,aRRange"                         */
   ,{"sw", "su"  }   /* SW "SWap" */
   ,{ "s", "s"   }   /* S "Soon,CeaSe,SiSter"                           */
   ,{"S@", "sho" }   /* TIO */
   ,{ "S", "sh"  }   /* SH "SHip,Sure,staTIon"                          */
   ,{"tS", "ch"  }   /* CH "CHurCH,maTCH,naTUre"                        */
   ,{"ts", "ts"  }   /* TS "TSai"                                       */
 //,{"tu", "chyu"}   /* TU "TUlip" */
   ,{"tti","tei" }   /* TTI "kiTTY"                                     */
   ,{ "t", "t"   }   /* T "Tea,TighT,buTTon"                            */
   ,{ "T", "s"   }   /* TH "THing,auTHor,paTH"                          */
   ,{"vaI","bi"  }   /* VAI "GraVItation"                               */
   ,{"vj", "vy"  }   /* V "View,heaVy,moVe"                             */
   ,{"vi", "bi"  }   /* V "View,heaVy,moVe"                             */
   ,{ "v", "v"   }   /* V "View,heaVy,moVe"                             */
   ,{"w3", "wa^" }   /* WO "WOrk" */
   ,{ "w", "u"   }   /* W "Wet,One,When,qUeen"                          */
   ,{"j0", "yo"  }   /* YO */
   ,{"ju", "yu^" }   /* YU */
   ,{"j@", "yo"  }   /* YO */
   ,{"j&", "ya"  }   /* YA */
   ,{"jV", "ya"  }   /* YA (YUp) */
   ,{ "j", "u"   }   /* Y "Yet,Use,bEAuty"                              */
   ,{"zI", "ti"  }   /* zI (?)                                          */
   ,{ "z", "z"   }   /* Z "Zero,Zone,roSeS,buZZ"                        */
   ,{ "Z", "j"   }   /* ZH "pleaSure,viSIon"                            */
};
    unsigned hintcount = sizeof(tab) / sizeof(tab[0]);

    /* Convert english to phonemes */
    string s = Phonemize(english);
    
    if(ShowPhonemes)
    	cout << s << endl;
    
    /* Convert phonemes to romaji */
    string res;
    unsigned b=s.size();
    for(unsigned a=0; a < b; )
    {
        for(unsigned h=0; h<hintcount; ++h)
        {
            unsigned c;
            for(c=0; tab[h][0][c]; ++c)
                if(s[a+c] != tab[h][0][c])
                    goto next;
            res += tab[h][1];
            a += c;
            goto ok;
        next: ;
        }
        res += ' ';
        ++a;
    ok:
        continue;
    }
    
    /* Fix romaji */
    string final;
    string spaces;
    for(unsigned a=0; a<res.size(); ++a)
    {
        if(res[a] == ' ')
        {
            spaces += ' ';
            continue;
        }
        
        /* Non-space, flush previous spaces */
        final += spaces;
        spaces = "";
        
        /* vowel, ok */
        if(res[a] == ' '
        || res[a] == '^'
        || res[a] == 'a'
        || res[a] == 'i'
        || res[a] == 'u'
        || res[a] == 'e'
        || res[a] == 'o'
        || res[a] == 'n'
        || res[a] == 'y')
        {
            final += res[a];
            continue;
        }
        /* double consonant, ok */
        if(res[a] == res[a+1])
        {
            final += res[a];
            continue;
        }
        /* consonant, not double */
        /* is it tsu, chi, shi? */
        if((res[a] == 't' && res[a+1] == 's')
        || (res[a] == 'c' && res[a+1] == 'h')
        || (res[a] == 's' && res[a+1] == 'h'))
        {
            /* yes, eat first char */
            final += res[a];
            ++a;
        }
        char base = res[a];
        if(base == 's' && res[a+1] == 'i')
        {
            final += 's';
            base = 'h';
        }
        else if(base == 't' && res[a+1] == 'i')
        {
            final += 'c';
            base = 'h';
        }
        else if(base == 't' && res[a+1] == 'u')
        {
            final += 't';
            base = 's';
        }
        else if(base == 'h' && res[a+1] == 'u')
        {
            base = 'f';
        }
        
        bool glue = true;
COMBINE:
        /* is the next one vowel? */
        if(res[a+1] == 'a'
        || res[a+1] == 'i'
        || res[a+1] == 'u'
        || res[a+1] == 'e'
        || res[a+1] == 'o'
        || res[a+1] == 'y')
        {
            /* yes, ok */
            final += base;
            continue;
        }
        /* no, we must add u or o */

        /* is this a doublable vowel? */
        bool doubled = false;
        if(base != 'n'
        && base != 'm'
        && base != 'g'
        && base != 'h'
        && base != 'r'
        && base != 's'
        && base != 'z'
        && final.size()
        && final[final.size()-1] != ' '
        && final[final.size()-1] != base)
        {
            doubled = true;
        }
        char extravowel = 'u';
        if(base == 't'
        || base == 'd')
            extravowel = 'o';
        
        /* is this the end? */
        bool end = false;
        if(glue && res[a+1] == ' ')
        {
        	/* Glue now, but not the next time */
        	glue = false;
        	++a;
        	goto COMBINE;
        }
        if(a+1 == res.size() || res[a+1] == ' ')
            end = true;
        if(doubled && (end || base == 'k'))
            final += base;
        /* now add the consonant and extra vowel */
        if(base == 'm' && !end)
            final += 'n';
        else
        {
            final += base;
            final += extravowel;
        }
    }
    cout << final << endl;
    return 0;
}

int main(int argc, char **argv)
{
    for(;;)
    {
        int option_index = 0;
        static struct option long_options[] =
        {
            {"help",     0, 0,0},
            {"version",  0, 0,0},
            {"phonemes", 0, 0,0},
            {0,0,0,0}
        };
        int c = getopt_long(argc, argv, "hVp", long_options, &option_index);
        if(c==-1) break;
        switch(c)
        {
            case 0:
                switch(option_index)
                {
                    case 0: goto arg_h;
                    case 1: goto arg_V;
                    case 2: goto arg_p;
                }
                break;
            case 'V': arg_V:
                cout << VERSION"\n";
                return 0;
            case 'h': arg_h:
		        cout << "deeng v"VERSION" - Copyright (C) 1992,2004 Bisqwit (http://iki.fi/bisqwit/)\n"
		                "Usage: deeng <english>\nTransforms the english text to romaji.\n";
		        cout << " --help, -h     This help\n"
		                " --version, -V  Displays version information\n"
		                " --phonemes, -p Shows the phonemestring\n"
		                "\n";
		        return 0;
		    case 'p': arg_p:
		        ShowPhonemes = true;
		}
	}
	while(optind < argc)
	    Romaji(argv[optind++]);
    return 0;
}
